/*************************************************************************
 * gzrecover - A program to recover data from corrupted gzip files
 *
 * Copyright (c) 2002-2003 Aaron M. Renn (arenn@urbanophile.com)
 * See README for licensing information - this program is free software
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <zlib.h>

/* Global contants */
#define DEFAULT_OUTBUF_SIZE 65535

const static char *optstring = "ho:sv";
const static char *usage = "Usage: gzrecover [-hsv] [-o <outfile>] <infile>";


/* Global Variables */
static int split_mode = 0;
static int verbose_mode = 0;
static int outfile_specified = 0;
static char *user_outname;
static size_t outbuf_size = DEFAULT_OUTBUF_SIZE;

/* Display usage string and exit */
void
show_usage(int exit_status)
{
  fprintf(stderr, "%s\n", usage);
  exit(exit_status); 
}

/* Check system call error status */
void
check_error(int badcode, int rc, const char* callname)
{
  if (rc == badcode)
    {
       perror(callname);
       exit(1);
    }
}

/* Open output file for writing */
int
open_outfile(char *infile)
{
  int ofd;
  char *outfile, *ptr;
  static int suffix = 1;

  /* Build the output file name */
  if (outfile_specified)
    check_error(0, outfile = (char *)malloc(strlen(user_outname) + 9), "malloc");
  else
    check_error(0, outfile = (char *)malloc(strlen(infile) + 25), "malloc"); 
    
  if (!outfile_specified) /* Strip of .gz unless user specified name */
   {
     ptr = strstr(infile, ".gz");
     if (ptr)
       *ptr = '\0'; /* Bad form to directly edit command line */ 

     ptr = strrchr(infile, '/'); /* Kill pathname */
     if (ptr)
       infile = ptr+1;
   }

  if (outfile_specified && split_mode)
    sprintf(outfile, "%s.%d", user_outname, suffix++);
  else if (outfile_specified)
    strcpy(outfile, user_outname);
  else if (split_mode)
    sprintf(outfile, "%s.recovered.%d", infile, suffix++);
  else
    sprintf(outfile, "%s.recovered", infile);

  /* Open it up */
  check_error(-1, ofd = open(outfile, O_RDWR | O_CREAT, S_IWUSR | S_IRUSR),
              "open");

  if (verbose_mode)
    fprintf(stdout, "Opened output file for writing: %s\n", outfile);

  return(ofd);
}

/* Initialize the zlib decompression engine */
void
init_zlib(z_stream *d_stream, char *buffer, size_t bufsize)
{
  int rc;

  memset(d_stream, 0, sizeof(z_stream));
  d_stream->next_in = buffer;
  d_stream->avail_in = bufsize;

  rc = inflateInit2(d_stream, -15); /* Don't ask why -15 - I don't know */
  if (rc != Z_OK)
    {
      perror("inflateInit2");
      exit(1);
    }
}

/* Skip gzip header stuff we don't care about */ 
void
skip_gzip_header(z_stream *d_stream)
{
  char flags;
  unsigned int len;

  flags = d_stream->next_in[3];
  d_stream->next_in += 10;
  d_stream->avail_in -= 10;

  if ((flags & 0x04) !=0) /* Extra field */
    {
      len = (unsigned int)*d_stream->next_in;
      len += ((unsigned int)*(d_stream->next_in)) << 8;
      d_stream->next_in += (2 + len);
      d_stream->avail_in -= (2 + len);
    }

  if ((flags & 0x08) != 0)  /* Orig Name */
    {
      while(*d_stream->next_in != 0)
        {   
          ++d_stream->next_in;
          --d_stream->avail_in;
        }
      ++d_stream->next_in;
      --d_stream->avail_in;
   }

  if ((flags & 0x10) != 0) /* Comment */
    while(*d_stream->next_in != 0)
      {   
         ++d_stream->next_in;
         --d_stream->avail_in;
      }

  if ((flags & 0x02) != 0) /* Head CRC */
    {
      d_stream->next_in += 2;
      d_stream->avail_in -= 2 ;
    }
}

/* Main program driver */
int
main(int argc, char **argv)
{
  int opt, rc, ifd, ofd, founderr=0, foundgood=0;
  off_t errpos=0, errinc=0;
  struct stat statbuf;
  size_t insize;
  char *infile, *inbuf, *outbuf;
  z_stream d_stream;

  /* Parse options */
  while ((opt = getopt(argc, argv, optstring)) != -1)
    {
      switch (opt)
        {
          case 'h':
            show_usage(0); 

          case 'o':
            user_outname = optarg;
            outfile_specified = 1;
            break;

          case 's':
            split_mode = 1;
            break;

          case 'v':
            verbose_mode = 1;
            break;

          default:
            show_usage(1);
        }
    }
 
  if (optind == argc)
    show_usage(1);
  infile = argv[optind];

  /* Open input file and memory map */
  check_error(-1, ifd = open(infile, O_RDONLY), "open");
  memset(&statbuf, 0, sizeof(statbuf));
  check_error(-1, fstat(ifd, &statbuf), "fstat");
  insize = statbuf.st_size;
  check_error(-1, inbuf = mmap(0, insize, PROT_READ, MAP_PRIVATE, ifd, 0), 
              "mmap");

  if (verbose_mode)
    fprintf(stdout, "Opened input file for reading: %s\n", infile);

  /* Open output file & initialize output buffer */
  ofd = open_outfile(infile);
  check_error(0, outbuf = (char *)malloc(outbuf_size), "malloc");

  /* Initialize zlib */
  init_zlib(&d_stream, inbuf, insize);
  skip_gzip_header(&d_stream);

  /* Finally - decompress this bad boy */
  for (;;)
    {
      d_stream.next_out = outbuf;
      d_stream.avail_out = outbuf_size; 

      rc = inflate(&d_stream, Z_NO_FLUSH);

      /* Here is the strategy. If we bomb, we reset zlib to one byte past the
       * error location and keep doing it until such time as we are able
       * to start decompressing something.  Alas, this seems to result in
       * a number of false starts.
       */ 
      if ((rc != Z_OK) && (rc != Z_STREAM_END))
        {
          foundgood = 0;
          if (!founderr)
            {
              errpos = insize - d_stream.avail_in;
              founderr = 1;

              if (d_stream.avail_in == 0)
                break;

              if (verbose_mode)
                fprintf(stdout, "Found error at byte %d in input stream\n",
                        errpos);
            }

          inflateEnd(&d_stream);
	  ++errinc;
          init_zlib(&d_stream, inbuf+errpos+errinc, insize - (errpos+errinc));

          continue;
        }

      if (founderr & !foundgood)
        {
          foundgood = 1;
          founderr = 0;
          errinc = 0;
	  if (verbose_mode)
            fprintf(stdout, "Found good data at byte %d in input stream\n",
                    errpos + errinc);

          if (split_mode)
            {
              close(ofd);
              ofd = open_outfile(infile);
            }
        }

      /* Write decompressed output - should really handle short write counts */
      check_error(-1, write(ofd, outbuf, outbuf_size - d_stream.avail_out),
                  "write");
      fsync(ofd);

      /* In we get a false alarm on end of file, we need to handle that to.
       * Reset to one byte past where it occurs */
      if (rc == Z_STREAM_END)
        {
          off_t tmpavail = d_stream.avail_in;

          if (tmpavail <= 64) /* Usually find some slop at end */
            break; 

          if (verbose_mode)
            fprintf(stdout, "Premature end of stream at %d\n", insize-tmpavail);

          inflateEnd(&d_stream);
          init_zlib(&d_stream, inbuf + (insize - tmpavail), tmpavail);
	  continue;
        }
    }

  /* Close up files */
  munmap(inbuf, insize);
  close(ofd);
  close(ifd);

  if (verbose_mode)
    fprintf(stdout, "Total decompressed output = %d bytes\n", 
            d_stream.total_out);

  return(0); 
}

